@@ -23,8 +23,6 @@ import (
 	"reflect"
 	"time"
 
-	"k8s.io/apimachinery/pkg/runtime/schema"
-
 	"github.com/go-logr/logr"
 	"github.com/pkg/errors"
 	"k8s.io/client-go/util/retry"
@@ -41,17 +39,20 @@ import (
 	corev1 "k8s.io/api/core/v1"
 	apierrors "k8s.io/apimachinery/pkg/api/errors"
 	"k8s.io/apimachinery/pkg/runtime"
+	"k8s.io/apimachinery/pkg/runtime/schema"
+	"k8s.io/apimachinery/pkg/types"
 	"k8s.io/apimachinery/pkg/util/wait"
 
 	nmstateapi "github.com/nmstate/kubernetes-nmstate/api/shared"
 	nmstatev1beta1 "github.com/nmstate/kubernetes-nmstate/api/v1beta1"
+	"github.com/nmstate/kubernetes-nmstate/pkg/enactment"
 	"github.com/nmstate/kubernetes-nmstate/pkg/enactmentstatus"
 	enactmentconditions "github.com/nmstate/kubernetes-nmstate/pkg/enactmentstatus/conditions"
 	"github.com/nmstate/kubernetes-nmstate/pkg/environment"
 	nmstate "github.com/nmstate/kubernetes-nmstate/pkg/helper"
+	"github.com/nmstate/kubernetes-nmstate/pkg/node"
 	"github.com/nmstate/kubernetes-nmstate/pkg/policyconditions"
 	"github.com/nmstate/kubernetes-nmstate/pkg/selectors"
-	"k8s.io/apimachinery/pkg/types"
 )
 
 var (
@@ -66,7 +67,7 @@ var (
 		},
 		UpdateFunc: func(updateEvent event.UpdateEvent) bool {
 			// [1] https://blog.openshift.com/kubernetes-operators-best-practices/
-			generationIsDifferent := updateEvent.MetaNew.GetGeneration() != updateEvent.MetaOld.GetGeneration()
+			generationIsDifferent := updateEvent.ObjectNew.GetGeneration() != updateEvent.ObjectOld.GetGeneration()
 			return generationIsDifferent
 		},
 	}
@@ -79,8 +80,8 @@ var (
 			return false
 		},
 		UpdateFunc: func(updateEvent event.UpdateEvent) bool {
-			labelsChanged := !reflect.DeepEqual(updateEvent.MetaOld.GetLabels(), updateEvent.MetaNew.GetLabels())
-			return labelsChanged && nmstate.EventIsForThisNode(updateEvent.MetaNew)
+			labelsChanged := !reflect.DeepEqual(updateEvent.ObjectOld.GetLabels(), updateEvent.ObjectNew.GetLabels())
+			return labelsChanged && nmstate.EventIsForThisNode(updateEvent.ObjectNew)
 		},
 		GenericFunc: func(event.GenericEvent) bool {
 			return false
@@ -88,6 +89,13 @@ var (
 	}
 )
 
+// NodeNetworkConfigurationPolicyReconciler reconciles a NodeNetworkConfigurationPolicy object
+type NodeNetworkConfigurationPolicyReconciler struct {
+	client.Client
+	Log    logr.Logger
+	Scheme *runtime.Scheme
+}
+
 func init() {
 	if !environment.IsHandler() {
 		return
@@ -99,108 +107,12 @@ func init() {
 	}
 }
 
-// NodeNetworkConfigurationPolicyReconciler reconciles a NodeNetworkConfigurationPolicy object
-type NodeNetworkConfigurationPolicyReconciler struct {
-	client.Client
-	Log    logr.Logger
-	Scheme *runtime.Scheme
-}
-
-func (r *NodeNetworkConfigurationPolicyReconciler) waitEnactmentCreated(enactmentKey types.NamespacedName) error {
-	var enactment nmstatev1beta1.NodeNetworkConfigurationEnactment
-	pollErr := wait.PollImmediate(1*time.Second, 10*time.Second, func() (bool, error) {
-		err := r.Client.Get(context.TODO(), enactmentKey, &enactment)
-		if err != nil {
-			if apierrors.IsNotFound(err) {
-				// Let's retry after a while, sometimes it takes some time
-				// for enactment to be created
-				return false, nil
-			}
-			return false, err
-		}
-		return true, nil
-	})
-
-	return pollErr
-}
-
-func (r *NodeNetworkConfigurationPolicyReconciler) initializeEnactment(policy nmstatev1beta1.NodeNetworkConfigurationPolicy) error {
-	enactmentKey := nmstateapi.EnactmentKey(nodeName, policy.Name)
-	log := r.Log.WithName("initializeEnactment").WithValues("policy", policy.Name, "enactment", enactmentKey.Name)
-	// Return if it's already initialize or we cannot retrieve it
-	enactment := nmstatev1beta1.NodeNetworkConfigurationEnactment{}
-	err := r.Client.Get(context.TODO(), enactmentKey, &enactment)
-	if err != nil && !apierrors.IsNotFound(err) {
-		return errors.Wrap(err, "failed getting enactment ")
-	}
-	if err != nil && apierrors.IsNotFound(err) {
-		log.Info("creating enactment")
-		enactment = nmstatev1beta1.NewEnactment(nodeName, policy)
-		err = r.Client.Create(context.TODO(), &enactment)
-		if err != nil {
-			return errors.Wrapf(err, "error creating NodeNetworkConfigurationEnactment: %+v", enactment)
-		}
-		err = r.waitEnactmentCreated(enactmentKey)
-		if err != nil {
-			return errors.Wrapf(err, "error waitting for NodeNetworkConfigurationEnactment: %+v", enactment)
-		}
-	} else {
-		enactmentConditions := enactmentconditions.New(r.Client, enactmentKey)
-		enactmentConditions.Reset()
-	}
-
-	return enactmentstatus.Update(r.Client, enactmentKey, func(status *nmstateapi.NodeNetworkConfigurationEnactmentStatus) {
-		status.DesiredState = policy.Spec.DesiredState
-		status.PolicyGeneration = policy.Generation
-	})
-}
-
-func (r *NodeNetworkConfigurationPolicyReconciler) enactmentsCountsByPolicy(policy *nmstatev1beta1.NodeNetworkConfigurationPolicy) (enactmentconditions.ConditionCount, error) {
-	enactments := nmstatev1beta1.NodeNetworkConfigurationEnactmentList{}
-	policyLabelFilter := client.MatchingLabels{nmstateapi.EnactmentPolicyLabel: policy.GetName()}
-	err := r.Client.List(context.TODO(), &enactments, policyLabelFilter)
-	if err != nil {
-		return nil, errors.Wrap(err, "getting enactment list failed")
-	}
-	enactmentCount := enactmentconditions.Count(enactments, policy.Generation)
-	return enactmentCount, nil
-}
-
-func (r *NodeNetworkConfigurationPolicyReconciler) claimNodeRunningUpdate(policy *nmstatev1beta1.NodeNetworkConfigurationPolicy) error {
-	policyKey := types.NamespacedName{Name: policy.GetName(), Namespace: policy.GetNamespace()}
-	err := r.Client.Get(context.TODO(), policyKey, policy)
-	if err != nil {
-		return err
-	}
-	if policy.Status.NodeRunningUpdate != "" && policy.Status.NodeRunningUpdate != nodeName {
-		return apierrors.NewConflict(schema.GroupResource{Resource: "nodenetworkconfigurationpolicies"}, policy.Name, fmt.Errorf("Another node is working on configuration"))
-	}
-	policy.Status.NodeRunningUpdate = nodeName
-	err = r.Client.Status().Update(context.TODO(), policy)
-	if err != nil {
-		return err
-	}
-	return nil
-}
-
-func (r *NodeNetworkConfigurationPolicyReconciler) releaseNodeRunningUpdate(policyKey types.NamespacedName) {
-	instance := &nmstatev1beta1.NodeNetworkConfigurationPolicy{}
-	_ = retry.RetryOnConflict(retry.DefaultRetry, func() error {
-		err := r.Client.Get(context.TODO(), policyKey, instance)
-		if err != nil {
-			return err
-		}
-		instance.Status.NodeRunningUpdate = ""
-		return r.Client.Status().Update(context.TODO(), instance)
-	})
-}
-
 // Reconcile reads that state of the cluster for a NodeNetworkConfigurationPolicy object and makes changes based on the state read
 // and what is in the NodeNetworkConfigurationPolicy.Spec
 // Note:
 // The Controller will requeue the Request to be processed again if the returned error is non-nil or
 // Result.Requeue is true, otherwise upon completion it will remove the work from the queue.
-func (r *NodeNetworkConfigurationPolicyReconciler) Reconcile(request ctrl.Request) (ctrl.Result, error) {
+func (r *NodeNetworkConfigurationPolicyReconciler) Reconcile(ctx context.Context, request ctrl.Request) (ctrl.Result, error) {
 	_ = context.Background()
 	log := r.Log.WithValues("nodenetworkconfigurationpolicy", request.NamespacedName)
 
@@ -221,7 +133,7 @@ func (r *NodeNetworkConfigurationPolicyReconciler) Reconcile(request ctrl.Reques
 
 	policyconditions.Reset(r.Client, request.NamespacedName)
 
-	err = r.initializeEnactment(*instance)
+	previousConditions, err := r.initializeEnactment(*instance)
 	if err != nil {
 		log.Error(err, "Error initializing enactment")
 	}
@@ -246,28 +158,29 @@ func (r *NodeNetworkConfigurationPolicyReconciler) Reconcile(request ctrl.Reques
 	}
 
 	enactmentConditions.NotifyMatching()
-	if !instance.Spec.Parallel {
-		enactmentCount, err := r.enactmentsCountsByPolicy(instance)
-		if err != nil {
-			log.Error(err, "Error getting enactment counts")
-			return ctrl.Result{}, nil
-		}
-		if enactmentCount.Failed() > 0 {
-			err = fmt.Errorf("policy has failing enactments, aborting")
-			log.Error(err, "")
-			enactmentConditions.NotifyAborted(err)
-			return ctrl.Result{}, nil
-		}
-		err = r.claimNodeRunningUpdate(instance)
+
+	enactmentCount, err := enactment.CountByPolicy(r.Client, instance)
+	if err != nil {
+		log.Error(err, "Error getting enactment counts")
+		return ctrl.Result{}, err
+	}
+	if enactmentCount.Failed() > 0 {
+		err = fmt.Errorf("policy has failing enactments, aborting")
+		log.Error(err, "")
+		enactmentConditions.NotifyAborted(err)
+		return ctrl.Result{}, nil
+	}
+
+	if r.shouldIncrementUnavailableNodeCount(previousConditions) {
+		err = r.incrementUnavailableNodeCount(instance)
 		if err != nil {
 			if apierrors.IsConflict(err) {
 				return ctrl.Result{RequeueAfter: nodeRunningUpdateRetryTime}, err
-			} else {
-				return ctrl.Result{}, err
 			}
+			return ctrl.Result{}, err
 		}
-		defer r.releaseNodeRunningUpdate(request.NamespacedName)
 	}
+	defer r.decrementUnavailableNodeCount(instance)
 
 	enactmentConditions.NotifyProgressing()
 	nmstateOutput, err := nmstate.ApplyDesiredState(r.Client, instance.Spec.DesiredState)
@@ -288,9 +201,8 @@ func (r *NodeNetworkConfigurationPolicyReconciler) Reconcile(request ctrl.Reques
 }
 
 func (r *NodeNetworkConfigurationPolicyReconciler) SetupWithManager(mgr ctrl.Manager) error {
-
-	allPolicies := handler.ToRequestsFunc(
-		func(handler.MapObject) []reconcile.Request {
+	allPolicies := handler.MapFunc(
+		func(client.Object) []reconcile.Request {
 			log := r.Log.WithName("allPolicies")
 			allPoliciesAsRequest := []reconcile.Request{}
 			policyList := nmstatev1beta1.NodeNetworkConfigurationPolicyList{}
@@ -321,7 +233,7 @@ func (r *NodeNetworkConfigurationPolicyReconciler) SetupWithManager(mgr ctrl.Man
 	// is not needed since all NNCPs are going to be Reconcile at node startup.
 	err = ctrl.NewControllerManagedBy(mgr).
 		For(&corev1.Node{}).
-		Watches(&source.Kind{Type: &corev1.Node{}}, &handler.EnqueueRequestsFromMapFunc{ToRequests: allPolicies}, builder.WithPredicates(onLabelsUpdatedForThisNode)).
+		Watches(&source.Kind{Type: &corev1.Node{}}, handler.EnqueueRequestsFromMapFunc(allPolicies), builder.WithPredicates(onLabelsUpdatedForThisNode)).
 		Complete(r)
 	if err != nil {
 		return errors.Wrap(err, "failed to add controller to NNCP Reconciler listening Node events")
@@ -329,6 +241,99 @@ func (r *NodeNetworkConfigurationPolicyReconciler) SetupWithManager(mgr ctrl.Man
 	return nil
 }
 
+func (r *NodeNetworkConfigurationPolicyReconciler) initializeEnactment(policy nmstatev1beta1.NodeNetworkConfigurationPolicy) (*nmstateapi.ConditionList, error) {
+	enactmentKey := nmstateapi.EnactmentKey(nodeName, policy.Name)
+	log := r.Log.WithName("initializeEnactment").WithValues("policy", policy.Name, "enactment", enactmentKey.Name)
+	// Return if it's already initialize or we cannot retrieve it
+	enactment := nmstatev1beta1.NodeNetworkConfigurationEnactment{}
+	err := r.Client.Get(context.TODO(), enactmentKey, &enactment)
+	if err != nil && !apierrors.IsNotFound(err) {
+		return nil, errors.Wrap(err, "failed getting enactment ")
+	}
+	if err != nil && apierrors.IsNotFound(err) {
+		log.Info("creating enactment")
+		enactment = nmstatev1beta1.NewEnactment(nodeName, policy)
+		err = r.Client.Create(context.TODO(), &enactment)
+		if err != nil {
+			return nil, errors.Wrapf(err, "error creating NodeNetworkConfigurationEnactment: %+v", enactment)
+		}
+		err = r.waitEnactmentCreated(enactmentKey)
+		if err != nil {
+			return nil, errors.Wrapf(err, "error waitting for NodeNetworkConfigurationEnactment: %+v", enactment)
+		}
+	} else {
+		enactmentConditions := enactmentconditions.New(r.Client, enactmentKey)
+		enactmentConditions.Reset()
+	}
+
+	return &enactment.Status.Conditions, enactmentstatus.Update(r.Client, enactmentKey, func(status *nmstateapi.NodeNetworkConfigurationEnactmentStatus) {
+		status.DesiredState = policy.Spec.DesiredState
+		status.PolicyGeneration = policy.Generation
+	})
+}
+
+func (r *NodeNetworkConfigurationPolicyReconciler) waitEnactmentCreated(enactmentKey types.NamespacedName) error {
+	var enactment nmstatev1beta1.NodeNetworkConfigurationEnactment
+	pollErr := wait.PollImmediate(1*time.Second, 10*time.Second, func() (bool, error) {
+		err := r.Client.Get(context.TODO(), enactmentKey, &enactment)
+		if err != nil {
+			if apierrors.IsNotFound(err) {
+				// Let's retry after a while, sometimes it takes some time
+				// for enactment to be created
+				return false, nil
+			}
+			return false, err
+		}
+		return true, nil
+	})
+
+	return pollErr
+}
+
+func (r *NodeNetworkConfigurationPolicyReconciler) shouldIncrementUnavailableNodeCount(conditions *nmstateapi.ConditionList) bool {
+	return !enactmentstatus.IsProgressing(conditions)
+}
+
+func (r *NodeNetworkConfigurationPolicyReconciler) incrementUnavailableNodeCount(policy *nmstatev1beta1.NodeNetworkConfigurationPolicy) error {
+	policyKey := types.NamespacedName{Name: policy.GetName(), Namespace: policy.GetNamespace()}
+	err := r.Client.Get(context.TODO(), policyKey, policy)
+	if err != nil {
+		return err
+	}
+	maxUnavailable, err := node.MaxUnavailableNodeCount(r.Client, policy)
+	if err != nil {
+		return err
+	}
+	if policy.Status.UnavailableNodeCount >= maxUnavailable {
+		return apierrors.NewConflict(schema.GroupResource{Resource: "nodenetworkconfigurationpolicies"}, policy.Name, fmt.Errorf("maximal number of %d nodes are already processing policy configuration", policy.Status.UnavailableNodeCount))
+	}
+	policy.Status.UnavailableNodeCount += 1
+	err = r.Client.Status().Update(context.TODO(), policy)
+	if err != nil {
+		return err
+	}
+	return nil
+}
+
+func (r *NodeNetworkConfigurationPolicyReconciler) decrementUnavailableNodeCount(policy *nmstatev1beta1.NodeNetworkConfigurationPolicy) {
+	policyKey := types.NamespacedName{Name: policy.GetName(), Namespace: policy.GetNamespace()}
+	instance := &nmstatev1beta1.NodeNetworkConfigurationPolicy{}
+	err := retry.RetryOnConflict(retry.DefaultRetry, func() error {
+		err := r.Client.Get(context.TODO(), policyKey, instance)
+		if err != nil {
+			return err
+		}
+		if instance.Status.UnavailableNodeCount <= 0 {
+			return fmt.Errorf("no unavailable nodes")
+		}
+		instance.Status.UnavailableNodeCount -= 1
+		return r.Client.Status().Update(context.TODO(), instance)
+	})
+	if err != nil {
+		r.Log.Error(err, "error decrementing unavailableNodeCount")
+	}
+}
+
 func (r *NodeNetworkConfigurationPolicyReconciler) forceNNSRefresh(name string) {
 	log := r.Log.WithName("forceNNSRefresh").WithValues("node", name)
 	log.Info("forcing NodeNetworkState refresh after NNCP applied")
